import logging
import optparse
import os
import signal
from authserv import app
from authserv import auth
from authserv import protocol
from authserv.ratelimit import *
from flask import Flask, request, abort, make_response


@blacklist_on_auth_failure(key_from_args(0), count=5, period=600, ttl=43200)
@blacklist_on_auth_failure(key_from_args(4), count=5, period=600, ttl=43200)
def _auth(username, service, password, otp_token, source_ip):
    user = app.userdb.get_user(username, service)
    if not user:
        return protocol.ERR_AUTHENTICATION_FAILURE
    return auth.authenticate(user, service, password, otp_token)


@app.route('/api/1/auth', methods=('POST',))
@ratelimit_http_request(key_from_request(header='HTTP_X_FORWARDED_FOR'),
                        count=10, period=60)
@ratelimit_http_request(key_from_request(param='username'),
                        count=10, period=60)
def do_auth():
    service = request.form.get('service')
    username = request.form.get('username')
    password = request.form.get('password')
    otp_token = request.form.get('otp')
    source_ip = request.form.get('source_ip')

    if not service or not username:
        abort(400)

    try:
        result = _auth(username, service, password, otp_token, source_ip)
    except Exception, e:
        app.logger.exception('Unexpected exception in authenticate()')
        abort(500)

    app.logger.info(
        'AUTH %s %s otp=%s %s',
        username, service, otp_token and 'y' or 'n', result)

    response = make_response(result)
    response.headers['Cache-Control'] = 'no-cache'
    response.headers['Content-Type'] = 'text/plain'
    response.headers['Expires'] = '-1'
    return response


def create_app(userdb=None, mc=None):
    app.config.from_envvar('APP_CONFIG', silent=True)

    if not userdb:
        from authserv import ldap_model
        userdb = ldap_model.UserDb(
            app.config['LDAP_SERVICE_MAP'],
            app.config.get('LDAP_URI', 'ldap://127.0.0.1:389'),
            app.config['LDAP_BIND_DN'],
            app.config['LDAP_BIND_PW'])
    app.userdb = userdb

    if not mc:
        try:
            import pylibmc
            mc = pylibmc.Client(
                app.config['MEMCACHE_ADDR'], binary=True)
        except ImportError:
            import memcache
            mc = memcache.Client(
                app.config['MEMCACHE_ADDR'], debug=0)
    app.memcache = mc

    return app


def run_werkzeug(addr, port, ssl_ca, ssl_cert, ssl_key, dh_params):
    ssl_ctx = None
    if ssl_ca and os.path.exists(ssl_ca):
        from authserv import openssl
        ssl_ctx = openssl.create_server_context(
            ssl_cert, ssl_key, ssl_ca, dhparams)
    logging.info('starting werkzeug server on %s:%d', addr, port)
    app.run(host=addr, port=port, use_reloader=False, ssl_context=ssl_ctx)


def run_gevent(addr, port, ssl_ca, ssl_cert, ssl_key, dh_params):
    from gevent.monkey import patch_all
    patch_all()
    from gevent.pywsgi import WSGIServer
    ssl_args = {}
    if ssl_ca and os.path.exists(ssl_ca):
        import ssl
        ssl_args = {
            'server_side': True,
            'certfile': ssl_cert,
            'keyfile': ssl_key,
            'ca_certs': ssl_ca,
            'cert_reqs': ssl.CERT_REQUIRED,
            'ssl_version': ssl.PROTOCOL_TLSv1,
        }
    logging.info('starting gevent server on %s:%d', addr, port)
    WSGIServer((addr, port), app.wsgi_app, **ssl_args).serve_forever()


def run(engines, addr, port, ssl_ca, ssl_cert, ssl_key, dh_params):
    if engines:
        engines = engines.split(',')
    else:
        engines = ['gevent', 'werkzeug']

    for e in engines:
        fn = globals().get('run_' + e, None)
        if not fn:
            logging.error('Unknown HTTP engine "%s"', e)
            continue
        try:
            return fn(addr, port, ssl_ca, ssl_cert, ssl_key, dh_params)
        except ImportError:
            pass

    logging.fatal('No HTTP engine available to run the server')


def main():
    parser = optparse.OptionParser()
    parser.add_option('--config',
                      help='Configuration file')
    parser.add_option('--addr', dest='addr', default='0.0.0.0',
                      help='Address to listen on (default: %default)')
    parser.add_option('--port', type='int', default=1616,
                      help='TCP port to listen on (default: %default)')
    parser.add_option('--engine', dest='engine',
                      help='HTTP engine to use (default: try gevent, then werkzeug)')
    parser.add_option('--ca', dest='ssl_ca',
                      default='/etc/ai/internal_ca.pem',
                      help='SSL CA certificate file (default: %default)')
    parser.add_option('--ssl-cert', dest='ssl_cert',
                      default='/etc/ai/localhost_internal.pem',
                      help='SSL client certificate file (default: %default)')
    parser.add_option('--ssl-key', dest='ssl_key',
                      default='/etc/ai/localhost_internal.key',
                      help='SSL client key file (default: %default)')
    parser.add_option('--dhparams', dest='dh_params',
                      default='/etc/ai/dhparams',
                      help='Diffie-Helmann parameters file')
    parser.add_option('--debug', action='store_true')

    opts, args = parser.parse_args()
    if len(args) != 0:
        parser.error('Too many arguments')

    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        handler = logging.handlers.SysLogHandler(
            address='/dev/log',
            facility=logging.handlers.SysLogHandler.LOG_DAEMON)
        handler.setLevel(logging.INFO)
        logging.getLogger().addHandler(handler)

    if opts.config:
        os.setenv('APP_CONFIG', opts.config)
    app.config.update({'DEBUG': opts.debug})

    def _stopall(signo, frame):
        logging.info('terminating with signal %d', signo)
        os._exit(0)
    signal.signal(signal.SIGINT, _stopall)
    signal.signal(signal.SIGTERM, _stopall)

    run(opts.engine, opts.addr, opts.port, opts.ssl_ca,
        opts.ssl_cert, opts.ssl_key, opts.dh_params)


if __name__ == '__main__':
    main()
